@@ -15,7 +15,7 @@ use std::{
 use prost::encoding::WireType;
 use prost_types::{
     field_descriptor_proto, DescriptorProto, EnumDescriptorProto, EnumValueDescriptorProto,
-    FieldDescriptorProto, OneofDescriptorProto,
+    FieldDescriptorProto, FileDescriptorProto, OneofDescriptorProto,
 };
 
 use crate::descriptor::{
@@ -37,25 +37,25 @@ pub(super) struct TypeMap {
 #[derive(Clone, PartialEq, Eq)]
 pub struct MessageDescriptor {
     file_set: FileDescriptor,
-    ty: TypeId,
+    index: u32,
 }
 
 struct MessageDescriptorInner {
     full_name: Box<str>,
-    parent: Option<TypeId>,
+    parent: ParentKind,
     is_map_entry: bool,
     fields: BTreeMap<u32, FieldDescriptorInner>,
     field_names: HashMap<Box<str>, u32>,
     field_json_names: HashMap<Box<str>, u32>,
     oneof_decls: Box<[OneofDescriptorInner]>,
-    extensions: Vec<usize>,
+    extensions: Vec<u32>,
 }
 
 /// A oneof field in a protobuf message.
 #[derive(Clone, PartialEq, Eq)]
 pub struct OneofDescriptor {
     message: MessageDescriptor,
-    index: usize,
+    index: u32,
 }
 
 struct OneofDescriptorInner {
@@ -88,13 +88,13 @@ struct FieldDescriptorInner {
 #[derive(Clone, PartialEq, Eq)]
 pub struct ExtensionDescriptor {
     file_set: FileDescriptor,
-    index: usize,
+    index: u32,
 }
 
 pub struct ExtensionDescriptorInner {
     field: FieldDescriptorInner,
     number: u32,
-    parent: Option<TypeId>,
+    parent: ParentKind,
     extendee: TypeId,
     json_name: Box<str>,
 }
@@ -103,22 +103,22 @@ pub struct ExtensionDescriptorInner {
 #[derive(Clone, PartialEq, Eq)]
 pub struct EnumDescriptor {
     file_set: FileDescriptor,
-    ty: TypeId,
+    index: u32,
 }
 
 struct EnumDescriptorInner {
     full_name: Box<str>,
-    parent: Option<TypeId>,
-    value_names: HashMap<Box<str>, usize>,
+    parent: ParentKind,
+    value_names: HashMap<Box<str>, u32>,
     values: Vec<EnumValueDescriptorInner>,
-    default_value: usize,
+    default_value: u32,
 }
 
 /// A value in a protobuf enum type.
 #[derive(Clone, PartialEq, Eq)]
 pub struct EnumValueDescriptor {
     parent: EnumDescriptor,
-    index: usize,
+    index: u32,
 }
 
 struct EnumValueDescriptorInner {
@@ -177,9 +177,19 @@ pub enum Cardinality {
     Repeated,
 }
 
+#[derive(Copy, Clone, Debug)]
+enum ParentKind {
+    File { index: u32 },
+    Message { index: u32 },
+}
+
 impl MessageDescriptor {
     pub(in crate::descriptor) fn new(file_set: FileDescriptor, ty: TypeId) -> Self {
-        MessageDescriptor { file_set, ty }
+        debug_assert_eq!(ty.0, field_descriptor_proto::Type::Message);
+        MessageDescriptor {
+            file_set,
+            index: ty.1,
+        }
     }
 
     pub(in crate::descriptor) fn iter(
@@ -189,10 +199,7 @@ impl MessageDescriptor {
             .inner
             .type_map
             .messages()
-            .map(move |ty| MessageDescriptor {
-                file_set: file_set.clone(),
-                ty,
-            })
+            .map(move |ty| MessageDescriptor::new(file_set.clone(), ty))
     }
 
     pub(in crate::descriptor) fn try_get_by_name(
@@ -203,10 +210,7 @@ impl MessageDescriptor {
         if !ty.is_message() {
             return None;
         }
-        Some(MessageDescriptor {
-            file_set: file_set.clone(),
-            ty,
-        })
+        Some(MessageDescriptor::new(file_set.clone(), ty))
     }
 
     /// Gets a reference to the [`FileDescriptor`] this message is defined in.
@@ -216,10 +220,10 @@ impl MessageDescriptor {
 
     /// Gets the parent message type if this message type is nested inside a another message, or `None` otherwise
     pub fn parent_message(&self) -> Option<MessageDescriptor> {
-        self.message_ty().parent.map(|ty| MessageDescriptor {
-            file_set: self.file_set.clone(),
-            ty,
-        })
+        self.message_ty()
+            .parent
+            .as_message()
+            .map(|ty| MessageDescriptor::new(self.file_set.clone(), ty))
     }
 
     /// Gets the short name of the message type, e.g. `MyMessage`.
@@ -236,13 +240,19 @@ impl MessageDescriptor {
     ///
     /// If no package name is set, an empty string is returned.
     pub fn package_name(&self) -> &str {
-        parse_namespace(&self.root_message_ty().full_name)
+        self.parent_file_descriptor_proto().package()
+    }
+
+    /// Gets a reference to the [`FileDescriptorProto`] in which this message is defined.
+    pub fn parent_file_descriptor_proto(&self) -> &FileDescriptorProto {
+        self.message_ty()
+            .parent
+            .resolve_parent_file_proto(&self.file_set)
     }
 
     /// Gets a reference to the raw [`DescriptorProto`] wrapped by this [`MessageDescriptor`].
     pub fn descriptor_proto(&self) -> &DescriptorProto {
-        self.parent_file()
-            .find_message_descriptor_proto(self.package_name(), self.full_name())
+        find_message_descriptor_proto(self.parent_file(), self.index)
     }
 
     /// Gets an iterator yielding a [`FieldDescriptor`] for each field defined in this message.
@@ -258,10 +268,8 @@ impl MessageDescriptor {
 
     /// Gets an iterator yielding a [`OneofDescriptor`] for each oneof field defined in this message.
     pub fn oneofs(&self) -> impl ExactSizeIterator<Item = OneofDescriptor> + '_ {
-        (0..self.message_ty().oneof_decls.len()).map(move |index| OneofDescriptor {
-            message: self.clone(),
-            index,
-        })
+        (0..self.message_ty().oneof_decls.len())
+            .map(move |index| OneofDescriptor::new(self.clone(), index))
     }
 
     /// Gets a [`FieldDescriptor`] with the given number, or `None` if no such field exists.
@@ -385,15 +393,7 @@ impl MessageDescriptor {
     }
 
     fn message_ty(&self) -> &MessageDescriptorInner {
-        self.file_set.inner.type_map.get_message(self.ty)
-    }
-
-    fn root_message_ty(&self) -> &MessageDescriptorInner {
-        let mut curr = self.message_ty();
-        while let Some(parent) = curr.parent {
-            curr = self.file_set.inner.type_map.get_message(parent);
-        }
-        curr
+        self.file_set.inner.type_map.get_message(self.index)
     }
 }
 
@@ -509,10 +509,7 @@ impl FieldDescriptor {
     pub fn containing_oneof(&self) -> Option<OneofDescriptor> {
         self.message_field_ty()
             .oneof_index
-            .map(|index| OneofDescriptor {
-                message: self.message.clone(),
-                index,
-            })
+            .map(|index| OneofDescriptor::new(self.message.clone(), index))
     }
 
     pub(crate) fn default_value(&self) -> Option<&crate::Value> {
@@ -561,7 +558,7 @@ impl ExtensionDescriptor {
             .extensions()
             .map(move |index| ExtensionDescriptor {
                 file_set: file_set.clone(),
-                index,
+                index: index.try_into().expect("index too large"),
             })
     }
 
@@ -575,10 +572,10 @@ impl ExtensionDescriptor {
     /// Note this just corresponds to where the extension was defined in the proto file. See [`containing_message`][ExtensionDescriptor::containing_message]
     /// for the message this field extends.
     pub fn parent_message(&self) -> Option<MessageDescriptor> {
-        self.extension_ty().parent.map(|ty| MessageDescriptor {
-            file_set: self.file_set.clone(),
-            ty,
-        })
+        self.extension_ty()
+            .parent
+            .as_message()
+            .map(|ty| MessageDescriptor::new(self.file_set.clone(), ty))
     }
 
     /// Gets the short name of the extension field type, e.g. `my_extension`.
@@ -597,27 +594,29 @@ impl ExtensionDescriptor {
     ///
     /// If no package name is set, an empty string is returned.
     pub fn package_name(&self) -> &str {
-        match self.root_message_ty() {
-            Some(message) => parse_namespace(&message.full_name),
-            None => parse_namespace(self.full_name()),
-        }
+        self.parent_file_descriptor_proto().package()
+    }
+
+    /// Gets a reference to the [`FileDescriptorProto`] in which this extension is defined.
+    pub fn parent_file_descriptor_proto(&self) -> &FileDescriptorProto {
+        self.extension_ty()
+            .parent
+            .resolve_parent_file_proto(&self.file_set)
     }
 
     /// Gets a reference to the raw [`FieldDescriptorProto`] wrapped by this [`ExtensionDescriptor`].
     pub fn field_descriptor_proto(&self) -> &FieldDescriptorProto {
         let name = self.name();
-        let package = self.package_name();
-        let namespace = parse_namespace(self.full_name());
-
-        if namespace == package {
-            self.parent_file()
-                .filter_file_descriptor_protos_by_package(package)
-                .flat_map(|file| file.extension.iter())
+        let parent = self.extension_ty().parent;
+        if let Some(parent_message) = parent.resolve_parent_message_proto(self.parent_file()) {
+            parent_message
+                .extension
+                .iter()
                 .find(|extension| extension.name() == name)
                 .expect("extension not found")
         } else {
-            self.parent_file()
-                .find_message_descriptor_proto(package, namespace)
+            parent
+                .resolve_parent_file_proto(self.parent_file())
                 .extension
                 .iter()
                 .find(|extension| extension.name() == name)
@@ -688,10 +687,7 @@ impl ExtensionDescriptor {
 
     /// Gets the containing message that this field extends.
     pub fn containing_message(&self) -> MessageDescriptor {
-        MessageDescriptor {
-            file_set: self.file_set.clone(),
-            ty: self.extension_ty().extendee,
-        }
+        MessageDescriptor::new(self.file_set.clone(), self.extension_ty().extendee)
     }
 
     pub(crate) fn default_value(&self) -> Option<&crate::Value> {
@@ -709,20 +705,6 @@ impl ExtensionDescriptor {
     fn extension_ty(&self) -> &ExtensionDescriptorInner {
         self.file_set.inner.type_map.get_extension(self.index)
     }
-
-    fn root_message_ty(&self) -> Option<&MessageDescriptorInner> {
-        match self.extension_ty().parent {
-            Some(mut curr) => loop {
-                let message = self.file_set.inner.type_map.get_message(curr);
-                if let Some(parent) = message.parent {
-                    curr = parent;
-                } else {
-                    return Some(message);
-                }
-            },
-            None => None,
-        }
-    }
 }
 
 impl fmt::Debug for ExtensionDescriptor {
@@ -809,6 +791,14 @@ impl fmt::Debug for Kind {
 }
 
 impl EnumDescriptor {
+    pub(in crate::descriptor) fn new(file_set: FileDescriptor, ty: TypeId) -> Self {
+        debug_assert_eq!(ty.0, field_descriptor_proto::Type::Enum);
+        EnumDescriptor {
+            file_set,
+            index: ty.1,
+        }
+    }
+
     pub(in crate::descriptor) fn iter(
         file_set: &FileDescriptor,
     ) -> impl ExactSizeIterator<Item = Self> + '_ {
@@ -816,10 +806,7 @@ impl EnumDescriptor {
             .inner
             .type_map
             .enums()
-            .map(move |ty| EnumDescriptor {
-                file_set: file_set.clone(),
-                ty,
-            })
+            .map(move |ty| EnumDescriptor::new(file_set.clone(), ty))
     }
 
     pub(in crate::descriptor) fn try_get_by_name(
@@ -830,10 +817,7 @@ impl EnumDescriptor {
         if !ty.is_enum() {
             return None;
         }
-        Some(EnumDescriptor {
-            file_set: file_set.clone(),
-            ty,
-        })
+        Some(EnumDescriptor::new(file_set.clone(), ty))
     }
 
     /// Gets a reference to the [`FileDescriptor`] this enum type is defined in.
@@ -843,10 +827,10 @@ impl EnumDescriptor {
 
     /// Gets the parent message type if this enum type is nested inside a another message, or `None` otherwise
     pub fn parent_message(&self) -> Option<MessageDescriptor> {
-        self.enum_ty().parent.map(|ty| MessageDescriptor {
-            file_set: self.file_set.clone(),
-            ty,
-        })
+        self.enum_ty()
+            .parent
+            .as_message()
+            .map(|ty| MessageDescriptor::new(self.file_set.clone(), ty))
     }
 
     /// Gets the short name of the enum type, e.g. `MyEnum`.
@@ -863,31 +847,33 @@ impl EnumDescriptor {
     ///
     /// If no package name is set, an empty string is returned.
     pub fn package_name(&self) -> &str {
-        match self.root_message_ty() {
-            Some(message) => parse_namespace(&message.full_name),
-            None => parse_namespace(self.full_name()),
-        }
+        self.parent_file_descriptor_proto().package()
+    }
+
+    /// Gets a reference to the [`FileDescriptorProto`] in which this enum is defined.
+    pub fn parent_file_descriptor_proto(&self) -> &FileDescriptorProto {
+        self.enum_ty()
+            .parent
+            .resolve_parent_file_proto(&self.file_set)
     }
 
     /// Gets a reference to the raw [`EnumDescriptorProto`] wrapped by this [`EnumDescriptor`].
     pub fn enum_descriptor_proto(&self) -> &EnumDescriptorProto {
         let name = self.name();
-        let package = self.package_name();
-        let namespace = parse_namespace(self.full_name());
-
-        if namespace == package {
-            self.parent_file()
-                .filter_file_descriptor_protos_by_package(package)
-                .flat_map(|file| file.enum_type.iter())
-                .find(|extension| extension.name() == name)
-                .expect("enum not found")
-        } else {
-            self.parent_file()
-                .find_message_descriptor_proto(package, namespace)
+        let parent = self.enum_ty().parent;
+        if let Some(parent_message) = parent.resolve_parent_message_proto(self.parent_file()) {
+            parent_message
                 .enum_type
                 .iter()
-                .find(|enum_ty| enum_ty.name() == name)
-                .expect("enum not found")
+                .find(|extension| extension.name() == name)
+                .expect("extension not found")
+        } else {
+            parent
+                .resolve_parent_file_proto(self.parent_file())
+                .enum_type
+                .iter()
+                .find(|extension| extension.name() == name)
+                .expect("extension not found")
         }
     }
 
@@ -920,20 +906,15 @@ impl EnumDescriptor {
             .values
             .binary_search_by_key(&number, |v| v.number)
         {
-            Ok(index) => Some(EnumValueDescriptor {
-                parent: self.clone(),
-                index,
-            }),
+            Ok(index) => Some(EnumValueDescriptor::new(self.clone(), index)),
             Err(_) => None,
         }
     }
 
     /// Gets an iterator yielding a [`EnumValueDescriptor`] for each value in this enum.
     pub fn values(&self) -> impl ExactSizeIterator<Item = EnumValueDescriptor> + '_ {
-        (0..self.enum_ty().values.len()).map(move |index| EnumValueDescriptor {
-            parent: self.clone(),
-            index,
-        })
+        (0..self.enum_ty().values.len())
+            .map(move |index| EnumValueDescriptor::new(self.clone(), index))
     }
 
     /// Gets an iterator over reserved value number ranges in this enum.
@@ -953,21 +934,7 @@ impl EnumDescriptor {
     }
 
     fn enum_ty(&self) -> &EnumDescriptorInner {
-        self.file_set.inner.type_map.get_enum(self.ty)
-    }
-
-    fn root_message_ty(&self) -> Option<&MessageDescriptorInner> {
-        match self.enum_ty().parent {
-            Some(mut curr) => loop {
-                let message = self.file_set.inner.type_map.get_message(curr);
-                if let Some(parent) = message.parent {
-                    curr = parent;
-                } else {
-                    return Some(message);
-                }
-            },
-            None => None,
-        }
+        self.file_set.inner.type_map.get_enum(self.index)
     }
 }
 
@@ -983,6 +950,13 @@ impl fmt::Debug for EnumDescriptor {
 }
 
 impl EnumValueDescriptor {
+    fn new(parent: EnumDescriptor, index: usize) -> EnumValueDescriptor {
+        EnumValueDescriptor {
+            parent,
+            index: index.try_into().expect("index too large"),
+        }
+    }
+
     /// Gets a reference to the [`FileDescriptor`] this enum value is defined in.
     pub fn parent_file(&self) -> &FileDescriptor {
         self.parent.parent_file()
@@ -1019,7 +993,7 @@ impl EnumValueDescriptor {
     }
 
     fn enum_value_ty(&self) -> &EnumValueDescriptorInner {
-        &self.parent.enum_ty().values[self.index]
+        &self.parent.enum_ty().values[self.index as usize]
     }
 }
 
@@ -1034,6 +1008,13 @@ impl fmt::Debug for EnumValueDescriptor {
 }
 
 impl OneofDescriptor {
+    fn new(message: MessageDescriptor, index: usize) -> Self {
+        OneofDescriptor {
+            message,
+            index: index.try_into().expect("index too large"),
+        }
+    }
+
     /// Gets a reference to the [`FileDescriptor`] this oneof is defined in.
     pub fn parent_file(&self) -> &FileDescriptor {
         self.message.parent_file()
@@ -1056,7 +1037,7 @@ impl OneofDescriptor {
 
     /// Gets a reference to the raw [`OneofDescriptorProto`] wrapped by this [`OneofDescriptor`].
     pub fn oneof_descriptor_proto(&self) -> &OneofDescriptorProto {
-        &self.parent_message().descriptor_proto().oneof_decl[self.index]
+        &self.parent_message().descriptor_proto().oneof_decl[self.index as usize]
     }
 
     /// Gets an iterator yielding a [`FieldDescriptor`] for each field of the parent message this oneof contains.
@@ -1071,7 +1052,7 @@ impl OneofDescriptor {
     }
 
     fn oneof_ty(&self) -> &OneofDescriptorInner {
-        &self.message.message_ty().oneof_decls[self.index]
+        &self.message.message_ty().oneof_decls[self.index as usize]
     }
 }
 
@@ -1148,9 +1129,8 @@ impl TypeMap {
         }
     }
 
-    fn get_message(&self, ty: TypeId) -> &MessageDescriptorInner {
-        debug_assert_eq!(ty.0, field_descriptor_proto::Type::Message);
-        &self.messages[ty.1 as usize]
+    fn get_message(&self, index: u32) -> &MessageDescriptorInner {
+        &self.messages[index as usize]
     }
 
     fn get_message_mut(&mut self, ty: TypeId) -> &mut MessageDescriptorInner {
@@ -1158,13 +1138,12 @@ impl TypeMap {
         &mut self.messages[ty.1 as usize]
     }
 
-    fn get_enum(&self, ty: TypeId) -> &EnumDescriptorInner {
-        debug_assert_eq!(ty.0, field_descriptor_proto::Type::Enum);
-        &self.enums[ty.1 as usize]
+    fn get_enum(&self, index: u32) -> &EnumDescriptorInner {
+        &self.enums[index as usize]
     }
 
-    fn get_extension(&self, index: usize) -> &ExtensionDescriptorInner {
-        &self.extensions[index]
+    fn get_extension(&self, index: u32) -> &ExtensionDescriptorInner {
+        &self.extensions[index as usize]
     }
 
     fn messages(&self) -> impl ExactSizeIterator<Item = TypeId> {
@@ -1252,16 +1231,69 @@ impl TypeId {
             field_descriptor_proto::Type::Sint64 => Kind::Sint64,
             field_descriptor_proto::Type::String => Kind::String,
             field_descriptor_proto::Type::Bytes => Kind::Bytes,
-            field_descriptor_proto::Type::Enum => Kind::Enum(EnumDescriptor {
-                file_set: file_set.clone(),
-                ty: self,
-            }),
+            field_descriptor_proto::Type::Enum => {
+                Kind::Enum(EnumDescriptor::new(file_set.clone(), self))
+            }
             field_descriptor_proto::Type::Group | field_descriptor_proto::Type::Message => {
-                Kind::Message(MessageDescriptor {
-                    file_set: file_set.clone(),
-                    ty: self,
-                })
+                Kind::Message(MessageDescriptor::new(file_set.clone(), self))
             }
         }
     }
 }
+
+impl ParentKind {
+    fn as_message(&self) -> Option<TypeId> {
+        match *self {
+            ParentKind::File { .. } => None,
+            ParentKind::Message { index } => {
+                Some(TypeId(field_descriptor_proto::Type::Message, index))
+            }
+        }
+    }
+
+    fn resolve_parent_file_proto<'a>(
+        &self,
+        file_set: &'a FileDescriptor,
+    ) -> &'a FileDescriptorProto {
+        let mut curr = *self;
+        loop {
+            match curr {
+                ParentKind::File { index } => {
+                    return &file_set.file_descriptor_set().file[index as usize]
+                }
+                ParentKind::Message { index } => {
+                    curr = file_set.inner.type_map.get_message(index).parent;
+                }
+            }
+        }
+    }
+
+    fn resolve_parent_message_proto<'a>(
+        &self,
+        file_set: &'a FileDescriptor,
+    ) -> Option<&'a DescriptorProto> {
+        match *self {
+            ParentKind::File { .. } => None,
+            ParentKind::Message { index } => Some(find_message_descriptor_proto(file_set, index)),
+        }
+    }
+}
+
+fn find_message_descriptor_proto(file_set: &FileDescriptor, index: u32) -> &DescriptorProto {
+    let message = file_set.inner.type_map.get_message(index);
+    match message.parent {
+        ParentKind::File { index: file_index } => file_set.file_descriptor_set().file
+            [file_index as usize]
+            .message_type
+            .iter()
+            .find(|ty| ty.name() == parse_name(&message.full_name))
+            .expect("message not found"),
+        ParentKind::Message {
+            index: parent_index,
+        } => find_message_descriptor_proto(file_set, parent_index)
+            .nested_type
+            .iter()
+            .find(|ty| ty.name() == parse_name(&message.full_name))
+            .expect("message not found"),
+    }
+}
